import matplotlib.pyplot as plt
import numpy as np




def hist(x,y, log=False, bins=0.1, modus="alpha"):
    """
    Plots a histogram of the anomaly scores x, where y is the ground truth
    log=False: if True, the y-axis is logarithmic
    bins=0.1: the number of bins. Also understands relative values to the length of x
    normed: if True, the histogram is normalized
    modus: if "alpha", the alpha channel is used for there to be two plots. if "next" there will be a peak next to a peak. If "stack" then the diagrams will be stacked. stack and next currently require same size of normals as of abnormals
    """

    mn=np.min(x)
    mx=np.max(x)
    if type(bins) is float:
        bins=int(len(x)*bins)

    x0=[xx for xx,yy in zip(x,y) if not yy]
    x1=[xx for xx,yy in zip(x,y) if yy]


    if modus=="alpha":
        plt.hist(x0, bins=bins, alpha=0.5, log=log, range=(mn,mx), label="normal")
        plt.hist(x1, bins=bins, alpha=0.5, log=log, range=(mn,mx), label="abnormal")
    else:

        x0=np.expand_dims(x0, axis=1)
        x1=np.expand_dims(x1, axis=1)
        x=np.concatenate((x0,x1), axis=1)
        plt.hist(x, bins=bins, log=log, alpha=0.5, label=('normal', 'abnormal'),range=(mn,mx), stacked=(modus=="stack"))


    plt.legend()

    return plt.show










from ..utils import run_algo, eval_func, test_algo, test_algo_n
from ..data.iterate import load_dataset
from ..stats import Stats, average_stats, negative_infinity, maximum_stats

import numpy as np
import json

from ..helper import adaptricallable


class Logger(object):

    def __init__(self, algos, algo_names=None, verbose=0, addfeat=False):

        assert type(algos) == dict or (not algo_names is None and type(algos)==list), "algos must be a list, or algo_names must be provided"

        if type(algos) == dict:
            self.algos = list(algos.values())
            self.algo_names = list(algos.keys())
        else:
            self.algos = algos
            self.algo_names = algo_names

        self.verbose = verbose

        self.datasets=[]
        self.results=[]
        self.addfeat=addfeat

    def add_algo(self, algo_name, algo_func, dic):
        self.algos.append(algo_func)
        self.algo_names.append(algo_name)
        for i, ds in enumerate(self.datasets):
            #assert ds.name() in dic.keys(), f"this algo is not complete. Missing {ds} got {dic.keys()}"
            if ds.name() in dic.keys():
                self.results[i].append(Stats(dic[ds.name()]))
            else:
                self.results[i].append(Stats(0.0))


    def add_run(self,dataset, results, verbose=None):
        if verbose is None:
            verbose = self.verbose
        
        if any([zw.isnan() for zw in results]):
            if verbose>0:
                print(f"Some results are NaN, skipping {dataset}")
            return
        self.datasets.append(dataset)
        self.results.append(results)

    def save(self, path):
        with open(path, 'w') as f:
            json.dump({'datasets':[zw.name() for zw in self.datasets], 'results':[[zw.to_list() for zw in zx] for zx in self.results],'algo_names':self.algo_names}, f, indent=2)

    def load(self,path):
        with open(path, 'r') as f:
            d = json.load(f)
        self.datasets = [load_dataset(zw) for zw in d['datasets']]
        self.results = [[Stats(zw) for zw in zx] for zx in d['results']]
        self.algo_names = d['algo_names']
        self.algos=[None for _ in self.algo_names]

    def run_on(self, d, x, tx, ty, n=10, verbose=None):
        if verbose is None:
            verbose = self.verbose
        if verbose>0:
            print("Running on dataset:", d)

        lis=[]
        for algo, nam in zip(self.algos, self.algo_names):

            results = test_algo_n(algo, x, tx, ty, n=n)
            if verbose>1:
                print(f"Results of {nam}: {results.shortstr()}")
            lis.append(results)
        self.add_run(d, lis,verbose=verbose)

    def run_cross(self, d, folds, verbose=None):
        if verbose is None:
            verbose = self.verbose
        if verbose>0:
            print("Running on dataset:", d)

        lis=[]
        for algo, nam in zip(self.algos, self.algo_names):
            results=[]
            for x,tx,ty in folds:
                qual = test_algo(algo, x, tx, ty)
                results.append(qual)
            results=Stats(results)
            if verbose>1:
                print(f"Results of {nam}: {results.shortstr()}")
            lis.append(results)
        self.add_run(d, lis, verbose=verbose)




    def run_on_all(self, iterable, *args, **kwargs):
        for zw in iterable:
            self.run_on(*zw, *args, **kwargs)

    def _sort(self, addmean=True):
        """helper function to sort stuff before showing it. Just to look nicer"""
        algo_names = self.algo_names
        datasets = self.datasets
        results = self.results
        means=np.array([[zw.mean() for zw in zx] for zx in results])
        i_algo_names = [i for i,zw in enumerate(algo_names)]
        i_datasets = [i for i,zw in enumerate(datasets)]

        score_dataset=np.mean(means, axis=1)
        score_algo=np.mean(means, axis=0)

        i_algo_names.sort(key=lambda x: -score_algo[x])
        i_datasets.sort(key=lambda x: score_dataset[x])

        algo_names = [algo_names[i] for i in i_algo_names]
        datasets = [datasets[i] for i in i_datasets]
        results = [[results[i][j] for j in i_algo_names] for i in i_datasets]

        if addmean:
            datasets.append("")
            datasets.append("Average")
            toa1=["" for _ in algo_names]
            toa2=[average_stats([zw[i] for zw in results]) for i in range(len(algo_names))]
            results.append(toa1)
            results.append(toa2)

        return algo_names, datasets, results



    def show(self, addfeat=None):
        if addfeat is None:
            addfeat=self.addfeat

        try:
            from tabulate import tabulate
        except ImportError:
            print("tabulate not installed, cannot show results!")
            return

        algo_names, datasets, results = self._sort()

        headers = ['Dataset'] + algo_names
        rows = []
        def trafo_to_str(q):
            return adaptricallable(q, "mediumstr")()

        for d, r in zip(datasets, results):
            rows.append([adaptricallable(d,"name_and_feat" if addfeat else "name")()] + [trafo_to_str(zw) for zw in r])
        print(tabulate(rows, headers=headers))


    def to_latex(self, addfeat=None):
        if addfeat is None:
            addfeat=self.addfeat
        try:
            from tabulate import tabulate
        except ImportError:
            print("tabulate not installed, cannot show results!")
            return
        algo_names, datasets, results = self._sort()
        headers = ['Dataset'] + algo_names
        rows = []
        vals = []
        def trafo_to_str(q):
            return adaptricallable(q, "mediumstr")()

        for d, r in zip(datasets, results):
            rows.append([adaptricallable(d,"name_and_feat" if addfeat else "name")()] + [trafo_to_str(zw) for zw in r])
            if type(r[0]) is str:
                vals.append([negative_infinity() for _ in range(len(r)+1)])
 
            else:    
                vals.append([negative_infinity()]+ [zw for zw in r])

        #print(vals)
        #print(len(vals))
        #print(len(vals[0]))
        #exit()

        maximas=[maximum_stats(zw) for zw in vals]
        maximas=[[trafo_to_str(zw) for zw in zx] for zx in maximas]

        def boldspace(s, keys=None):
            s=s.strip()
            if keys is None:
                keys = ["+"]
            key=keys[0]
            if key in s:
                bef=s[:s.index(key)]
                aft=s[s.index(key):]
                s=f"\\textbf{{{bef}}} {aft}"
            
            if len(keys)>1:
                return boldspace(s, keys=keys[1:])
            else:
                if "\\textbf" in s:
                    return s
                else:
                    return f"\\textbf{{{s}}}"

        def latexify(s, bold=False):
            if bold:
                return latexify(boldspace(s))
            return "$"+str(s).replace("_","\_").replace("+-"," \\pm ")+"$"#not the best solution
        rows = [[latexify(zw, zw in maxima) for zw in zx] for zx,maxima in zip(rows,maximas)]
        return tabulate(rows, headers=headers, tablefmt="latex_raw")


















import numpy as np
from ..helper import seed_once


def remove_const_features(x):
    """
    Remove constant features from a numpy array.
    """
    assert len(x.shape) == 2, "remove_const_features only works on 2D arrays"
    return x[:, np.where(np.std(x, axis=0) > 0)[0]]

@seed_once
def train_test_split(x,y,max_test_size=0.3):
    """
    Split data into training and test sets.
    """
    assert len(x.shape) == 2, "train_test_split only works on 2D arrays"
    norm=np.array([xx for xx,yy in zip(x,y) if yy==0])
    anom=np.array([xx for xx,yy in zip(x,y) if yy!=0])
    
    np.random.shuffle(norm)
    np.random.shuffle(anom)

    n_ano=int(anom.shape[0])
    n_ano=min([int(max_test_size*norm.shape[0]),n_ano])

    train=norm[:-n_ano]
    testn=norm[-n_ano:]
    testa=anom[:n_ano]
    test=np.concatenate((testn,testa),axis=0)
    testy=np.array([0]*n_ano+[1]*n_ano)

    return train,test,testy

@seed_once
def shuffle(x,y=None):
    """
    Shuffle the data. Keeps the connection between the data and the labels if provided
    """
    if y is not None:
        seed = np.random.randint(0, 2**32 - 1)
        np.random.seed(seed)
        np.random.shuffle(x)
        np.random.seed(seed)
        np.random.shuffle(y)
        return x,y
    else:
        np.random.shuffle(x)
        return x


@seed_once
def crossassign(n_samples,n_folds):
    """
    Randomly assigns samples to folds.
    """
    n_samples = int(n_samples)
    n_folds = int(n_folds)
    if n_samples < n_folds:
        raise ValueError("n_samples must be greater than n_folds")
    if n_folds < 2:
        raise ValueError("n_folds must be greater than 1")
    assigns=[i%n_folds for i in range(n_samples)]
    np.random.shuffle(assigns)
    return assigns

@seed_once
def crossvalidate(x,y,n_folds=5):
    """
    Splits the data into crossvalidation folds. Return n_folds versions of train, test and testy
    """
    x0=np.array([xx for xx,yy in zip(x,y) if yy==0])
    x1=np.array([xx for xx,yy in zip(x,y) if yy!=0])

    assigns=crossassign(len(x0),n_folds)

    def fold(x,y,i):    
        train=np.array([xx for ass,xx in zip(assigns,x0) if ass!=i])#training folds
        test_n=np.array([xx for ass,xx in zip(assigns,x0) if ass==i])#test fold
        if len(test_n)>=len(x1):#more normal data than abnormal one in the testing data. 
            test_n=test_n[:len(test_n)]#So cut the normal one
            x1m=x1#but keep the abnormal one
        elif len(test_n)<len(x1):#more abnormal data than normal one in the testing data.
            #folding so other abnormal samples are used sometime
            border=int((i/n_folds)*len(x1))
            a,b=x1[:border],x1[border:]
            x1m=np.concatenate((a,b),axis=0)
            x1m=x1m[:len(test_n)]
            #keep the testing one
        test=np.concatenate([test_n,x1m],axis=0)
        test_y=np.array([0]*len(test_n)+[1]*len(x1m))
        return train,test,test_y

    trains,tests,test_ys=[],[],[]
    for i in range(n_folds):
        train,test,test_y=fold(x,y,i)
        yield train,test,test_y

def normalize(x,sec=None,method="zscore"):
    """
    Normalize the data by the given method.
    if sec is provided, it is also normalized using the same method and statistics from the first one
    """

    if method=="zscore":
        xm=np.mean(x,axis=0)
        xs=np.std(x,axis=0)
        xs=np.where(xs==0,1,xs)
    elif method=="minmax":
        xm=np.min(x,axis=0)
        xs=np.max(x,axis=0)-xm
        xs=np.where(xs==0,1,xs)
    else:
        raise ValueError("Unknown normalization method: "+method)
    x=(x-xm)/xs
    if sec is not None:
        sec=(sec-xm)/xs
        return x,sec
    return x

def asfloat(x):
    return np.array(x).astype(np.float32)

def asint(x):
    return np.array(x).astype(np.int32)

def drop(x,t,what):
    if not type(what) == list:
        what=[what]
    lines=np.stack([xx for xx,tt in zip(x.T,t) if not tt in what],axis=1)
    t=[tt for tt in t if not tt in what]
    return lines,t

def convert_if_possible(x,t):
    if "categorical" in t:return x
    return asfloat(x)









class negate_output(object):

    def __init__(self, obj):
        self.obj = obj

    def fit(self, *args, **kwargs):
        return self.obj.fit(*args, **kwargs)

    def score_samples(self, *args, **kwargs):
        if hasattr(self.obj, 'score_samples'):
            return -self.obj.score_samples(*args, **kwargs)
        else:
            return -self.obj.decision_function(*args, **kwargs)







from .handler import adaptive
from ..helper import lambify

def dxy(iterable):
    """instead of yielding datasets, this iterator yields dataset, x, y"""
    return adaptive(iterable,"dxy")

def nonconst(iterable):
    """instead of yielding datasets, this iterator yields datasets without constant features"""
    return adaptive(iterable,"nonconst")

@lambify
def split(iterable,*args,**kwargs):
    """handles train test split for each input. Thus returns d, train, test_x, test_y"""
    return adaptive(iterable,"split",*args,**kwargs)

@lambify
def shuffle(iterable,seed=None):
    """shuffles whatever you put into it."""
    if seed is not None:
        return adaptive(iterable,"shuffle",seed=seed)
    else:
        return adaptive(iterable,"shuffle")

@lambify
def drop(iterable, what="categorical",*args,**kwargs):
    return adaptive(iterable,"drop",what,*args,**kwargs)

@lambify
def crossval(iterable, *args, **kwargs):
    """handles cross validation for each input. Thus returns d, train, test_x, test_y, but each n_folds times"""
    return adaptive(iterable,"crossval",*args,**kwargs)

def pipeline(iterable, *tasks):
    """pipeline is a way to chain multiple iterables together."""
    q=iterable
    for task in tasks:
        q=task(q)
    return q

@lambify
def normalize(iterable,method="zscore"):
    """normalizes the dataset"""
    return adaptive(iterable,"normalize_"+method)

def zscore(iterable):
    """normalizes the dataset so that the mean is 0 and the standard deviation is 1"""
    return normalize(iterable,"zscore")
normalize_zscore=zscore

def minmax(iterable):
    """normalizes the dataset so that max is 1 and min is 0"""
    return normalize(iterable,"minmax")
normalize_minmax=minmax

def listfold(iterable):
    return adaptive(iterable,"listfold")

def asfloat(iterable):
    return adaptive(iterable,"asfloat")

def asint(iterable):
    return adaptive(iterable,"asint")

@lambify
def cut(iterable, maxima=10):
    return adaptive(iterable,"cut",maxima=maxima)

def addtypes(iterable):
    return adaptive(iterable,"addt")

def dont_convert(iterable):
    return adaptive(iterable,"dontconvert")






import json
import numpy as np


from .methods import sign_test

from ..const import significance

class Stats(object):
    def __init__(self,q):
        self.q = q

    def __str__(self):
        if len(self.q)==0:
            return "No data"
        elif len(self.q)==1:
            return str(self.q[0])
        else:
            return f"{np.mean(self.q)} +- {np.std(self.q)} [{np.min(self.q)}, {np.max(self.q)}] * "+"{"+str(len(self.q))+"}"

    def __repr__(self):
        return str(self)

    def __float__(self):
        if len(self.q)==0:
            raise ValueError("No data")
        return np.mean(self.q)

    def __add__(self,other):
        assert isinstance(other,Stats)
        return Stats(self.q+other.q)

    def __eq__(self,other):
        assert isinstance(other,Stats)
        return sign_test(self.q,other.q)

    def __ne__(self,other):
        return 1-(self==other)

    def isnan(self):
        return np.any(np.isnan(self.q))

    def shortstr(self):
        return str(round(np.mean(self.q),4))

    def mediumstr(self):
        std=self.std()
        if std<0.0001:
            return str(round(self.mean(),4))
        return f"{self.mean():.4f}+-{self.std():.4f}"

    def to_list(self):
        return [float(x) for x in self.q]

    def mean(self):
        return np.mean(self.q)

    def std(self):
        return np.std(self.q)

    def __len__(self):
        return len(self.q)



if __name__=="__main__":
    q=Stats([1,2,3,4,5])
    print(q)


    exit()

def negative_infinity(n=10):
    return Stats([-2**30]*n)


def combine_stats(*args):
    assert len(args)>0
    assert all(isinstance(x,Stats) for x in args)

    ret=args[0]
    for x in args[1:]:
        ret=ret+x
    return ret

def average_stats(*args):
    """This is a simplistic averaging function. You could do better"""
    assert len(args)>0
    if len(args)==1:
        if type(args[0])==list:
            return average_stats(*args[0])
    assert all(isinstance(x,Stats) for x in args)

    ret=[]
    for x in args:
        ret.append(x.mean())
    return Stats(ret)

def maximum_stats(*args):
    """from a list of stats, which are the bests in them? (returns also a list, as in later versions we could implement statistical significance here. But currently only returns max(mean)"""

    assert len(args)>0
    if len(args)==1:
        if type(args[0])==list:
            return maximum_stats(*args[0])
    assert all(isinstance(x,Stats) for x in args)


    maxv=-2**30-1
    maxi=0
    for i,arg in enumerate(args):
        if arg.mean()>maxv:
            maxv=arg.mean()
            maxi=i
    maxo=args[maxi]
    return [arg for arg in args if len(maxo)==len(arg) and (maxo==arg)>significance]
    return [args[maxi]]






from sklearn.metrics import roc_auc_score
from copy import copy

from ..helper import seed_once
from ..stats import Stats


def run_algo_sklearn(algo, train, seed=42):
    algo=copy(algo)
    algo.random_state=seed
    algo.fit(train)
    return lambda x,model=algo: model.score_samples(x)

def run_algo_pyod(algo, train, seed=42):
    algo=copy(algo)
    algo.random_state=seed
    algo.fit(train)
    return lambda x,model=algo: model.decision_function(x)

@seed_once
def run_algo_callable(algo, train, *args, **kwargs):
    return algo(train,*args,**kwargs)





def run_algo(algo, train, *args,seed=42, **kwargs):
    if callable(algo):
        return run_algo_callable(algo, train, *args,seed=seed, **kwargs)
    if hasattr(algo, 'fit') and hasattr(algo, 'score_samples'):
        return run_algo_sklearn(algo, train, seed=seed)
    if hasattr(algo, 'fit') and hasattr(algo, 'decision_function'):
        return run_algo_pyod(algo, train, seed=seed)

def eval_func(func, test, testy, method="roc_auc"):
    if type(func) is list:
        preds=[]
        for f in func:
            preds.append(f(test))
        return Stats(pred)
    pred=func(test)
    if method=="roc_auc":
        return roc_auc_score(testy, pred)
    else:
        raise ValueError("Unknown method: {}".format(method))

def test_algo(algo, train, test, testy, *args, method="roc_auc",seed=42, **kwargs):
    func=run_algo(algo, train, *args,seed=seed, **kwargs)
    pred=func(test)
    return eval_func(func, test, testy)

def test_algo_n(algo, train, test, testy, *args, method="roc_auc", n=10, **kwargs):
    scores=[]
    for i in range(n):
        scores.append(test_algo(algo, train, test, testy, *args, method=method, seed=i, **kwargs))
    return Stats(scores)

def run_algo_folds(algo, folds, *args, seed=42, **kwargs):
    funcs=[]
    for x,tx,ty in folds:
        funcs.append(run_algo(algo, x, *args, seed=seed, **kwargs))
    return funcs

def run_algo_n(algo, train, *args, n=10, **kwargs):
    funcs=[]
    for i in range(n):
        funcs.append(run_algo(algo, train, *args, seed=i, **kwargs))
    return funcs












import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import backend as K

import os
import numpy as np
import time

from sklearn.metrics import roc_auc_score

from tqdm import tqdm

class DEAN(object):
    

    def __init__(s,task,data,test_data=None, test_labels=None,lr=0.03,batch=100,depth=3,bag=128,pth="./DEAN/",
            act="relu",reg=None,use_bias=False,goal_mean=1.0,verbose=1,patience=5,max_epochs=500,validation_split=0.2):

        s.task = task
        s.data = data
        s.test_data = test_data
        s.test_labels = test_labels
        s.lr = lr
        s.batch = batch
        s.depth = depth
        s.bag = bag
        s.pth = pth+task+"/"
        os.makedirs(s.pth,exist_ok=True)
        s.act = act
        s.reg = reg
        s.use_bias = use_bias
        s.goal_mean = goal_mean
        s.verbose = verbose
        s.patience = patience
        s.max_epochs = max_epochs
        s.validation_split = validation_split


    def build_one_model(s,index=0):
        np.random.seed(index)
        tf.random.set_seed(index)

        inn=keras.layers.Input(shape=(s.bag,))
        q=inn

        dims=[s.bag for i in range(s.depth)]
        for aq in dims[1:-1]:
            q=keras.layers.Dense(aq,activation=s.act,use_bias=s.use_bias,
                                  kernel_regularizer=s.reg)(q)
        q=keras.layers.Dense(dims[-1],activation="linear",use_bias=s.use_bias,
                                kernel_regularizer=s.reg)(q)
    
        model=keras.Model(inn,q)

        comp=K.ones_like(q)*s.goal_mean
        loss=keras.losses.msle(q,comp)
        loss=K.mean(loss)
        model.add_loss(loss)

        model.compile(optimizer=keras.optimizers.Adam(s.lr))
        return model

    def data_dex(s,index=0):
        ps=np.random.randint(0,2**31)
        np.random.seed(index)
        to_use=np.random.choice(s.data.shape[1],s.bag,replace=False)
        np.random.seed(ps)
        return to_use

    def data_trafo(s,data,to_use=None):
        if type(to_use) is int:
            to_use=s.data_dex(to_use)
        assert not to_use is None
        return np.concatenate([np.expand_dims(data[:,use],axis=1) for use in to_use],axis=1)

    def sub_data(s,index=0):
        to_use=s.data_dex(index)
        return s.data_trafo(s.data,to_use)

    def _seed(s,index=0):
        np.random.seed(index)
        tf.random.set_seed(index)
    
    def index_path(s,index=0):
        pth=f"{s.pth}/{index}/"
        os.makedirs(pth,exist_ok=True)
        return pth

    def train_one(s,index=0):
        pth=s.index_path(index)
        s._seed(index)

        to_use=s.data_dex(index)
        data=s.data_trafo(s.data,to_use)

        model=s.build_one_model(index)

        if s.verbose:
            model.summary()

        cb=[keras.callbacks.EarlyStopping(monitor="val_loss",patience=s.patience,verbose=s.verbose,restore_best_weights=True)]
        cb.append(keras.callbacks.TerminateOnNaN())
        cb.append(keras.callbacks.ModelCheckpoint(pth+"model.h5",monitor="val_loss",verbose=s.verbose,save_best_only=True,save_weights_only=True))

        t1=time.perf_counter()

        model.fit(data,None,
                epochs=s.max_epochs,
                batch_size=s.batch,
                validation_split=s.validation_split,
                verbose=s.verbose,
                callbacks=cb)

        t2=time.perf_counter()

        p=model.predict(data)
        m=np.mean(p)

        scores=[]
        auc=-1.0
        if not s.test_data is None:
            test_data=s.data_trafo(s.test_data,to_use)
            test_p=model.predict(test_data)
            test_p=np.mean(test_p,axis=-1)
            scores=(test_p-m)**2
            if not s.test_labels is None:
                auc=roc_auc_score(s.test_labels,scores)

        dt=t2-t1

        if s.verbose:
            print(f"{index} {auc:.4f} {dt:.4f}")

        np.savez_compressed(pth+"results.npz",m=m,s=scores,auc=auc,dt=dt,to_use=to_use)



        return scores

    def ensemble_merge(s,scores):
        return np.sqrt(np.mean(np.array(scores)**2,axis=0))

    def ensemble_eval(s,scores):
        assert not s.test_labels is None
        return roc_auc_score(s.test_labels,s.ensemble_merge(scores))

    def train_many(s,i0=0,i1=100):
        scores=[]
        aucs=[]
        tq=tqdm(range(i0,i1))
        for i in tq:
            scores.append(s.train_one(i))
            aucs.append(s.ensemble_eval(scores))
            tq.set_description(f"{i} {aucs[-1]:.4f}")

        return scores,aucs


















from .helper import remove_const_features, train_test_split, shuffle, crossvalidate, normalize, asfloat, asint, drop, convert_if_possible


class handler(object):
    def __init__(self, child, mods=None):
        self.child = child
        if mods is None:
            mods = []
        self.mods = mods
        self.args={}
        self.order=["d","x","y","train","testx","testy","folds","t"]


    def assertmod(self, mod):
        if mod not in self.mods:
            self.mods.append(mod)
    def fixmods(self):
        for zw in ["split","nonconst","split","crossval","normalize_zscore","normalize_minmax", "asfloat", "asint", "cut"]:
            if zw in self.mods:
                self.assertmod("dxy")
        if "listfold" in self.mods:
            self.assertmod("crossval")
        assert not ("crossval" in self.mods and "split" in self.mods), "test-train split and cross-validation are not mutually exclusive"

    def yield_order(self, q):
        ret=[]
        for key in self.order:
            if key in q:
                ret.append(q[key])
        return ret

    def __iter__(self):
        self.fixmods()
        for i,d in enumerate(self.child):
            if "cut" in self.mods:
                cut=self.args["cut"]["kwargs"]["maxima"]
                if i>=cut:break

            ac={"d":d,"t":d.types()}
            if "dxy" in self.mods:
                ac["x"],ac["y"]=ac["d"].getxy()

            if "drop" in self.mods:
                ac["x"],ac["t"]=drop(ac["x"],ac["t"],*self.args["drop"]["args"],**self.args["drop"]["kwargs"])

            if not "dontconvert" in self.mods:
                ac["x"]=convert_if_possible(ac["x"],ac["t"])



            if "asint" in self.mods:
                ac["x"]=asint(ac["x"])
                ac["y"]=asint(ac["y"])

            if "asfloat" in self.mods:
                ac["x"]=asfloat(ac["x"])
                ac["y"]=asfloat(ac["y"])

            if "nonconst" in self.mods:
                ac["x"]=remove_const_features(ac["x"])

            if "split" in self.mods:
                ac["train"],ac["testx"],ac["testy"]=train_test_split(ac["x"],ac["y"],*self.args["split"]["args"],**self.args["split"]["kwargs"])
                del ac["x"]
                del ac["y"]

            if "crossval" in self.mods:
                ac["folds"]=crossvalidate(ac["x"],ac["y"],*self.args["crossval"]["args"],**self.args["crossval"]["kwargs"])
                del ac["x"]
                del ac["y"]

            if "shuffle" in self.mods:
                try:
                    seed=self.args["shuffle"]["kwargs"]["seed"]
                    callab=lambda x,y=None: shuffle(x,y,seed=seed)
                except KeyError:
                    callab=shuffle
                
                if "split" in self.mods:
                    ac["train"]=callab(ac["train"])
                    ac["testx"],ac["testy"]=callab(ac["testx"],ac["testy"])
                elif "crossval" in self.mods:
                    ac["folds"]=((callab(x),*callab(tx,ty)) for x,tx,ty in ac["folds"])

                else:
                    ac["x"],ac["y"]=shuffle(ac["x"],ac["y"],seed=self.args["shuffle"]["kwargs"]["seed"])

            if "normalize_zscore" in self.mods:
                if "split" in self.mods:
                    ac["train"],ac["testx"]=normalize(ac["train"],ac["testx"],"zscore")
                elif "crossval" in self.mods:
                    ac["folds"]=((*normalize(x,tx,"zscore"),ty) for x,tx,ty in ac["folds"])
                else:
                    ac["x"]=normalize(ac["x"],method="zscore")
            if "normalize_minmax" in self.mods:
                if "split" in self.mods:
                    ac["train"],ac["testx"]=normalize(ac["train"],ac["testx"],"minmax")
                elif "crossval" in self.mods:
                    ac["folds"]=((*normalize(x,tx,"minmax"),ty) for x,tx,ty in ac["folds"])
                else:
                    ac["x"]=normalize(ac["x"],method="minmax")

            if not "addt" in self.mods:
                del ac["t"]


            yield self.yield_order(ac)

def arg_dic(*args,**kwargs):
    return {"args":[*args],"kwargs":{**kwargs}}

def adaptive(zw, mod, *args, **kwargs):
    if type(zw) is handler:
        zw.mods.append(mod)
        zw.args[mod]=arg_dic(*args,**kwargs)
        return zw
    else:
        ret= handler(zw, [mod])
        ret.args[mod]=arg_dic(*args,**kwargs)
        return ret







from datetime import datetime, timezone, timedelta


def timestamp():
    t=datetime.now(timezone.utc)
    x=datetime.timestamp(t)-12*3600
    x=round(x)
    return x



if __name__ == '__main__':
    print(timestamp())





from scipy import stats

def sign_test(a,b):
    """
    Performs a sign test on the data.
    """
    assert len(a) == len(b), "only the same number of elements can be compared"
    if a==b:return 1.0
    return stats.wilcoxon(a,b).pvalue






import numpy as np



def genmat(n,m):
    """
    Generate a random connection matrix of size n x m, where each connection is drawn with replacement
    Each element is 1 as long as not no try draws the connection
    So the probability of a connection is:
    prob_drawn_one_try=1/dim
    prob_drawn_dim_tries=1-(1-1/dim)^dim

    """
    dim=n*m
    prob=1-(1-1/dim)**dim

    mat=np.random.uniform(0,1,(n,m))
    return (mat<prob).astype(float)


if __name__ == "__main__":

    n=7
    m=7
    mat=genmat(n,m)
    from plt import plt
    plt.imshow(mat)
    plt.show()




import numpy as np
import scipy.stats as st

def to_likelihood(z):
    """convert z_score to likelyhood"""
    try:
        return [to_likelihood(x) for x in z]
    except TypeError:
        return 2*(1-st.norm.cdf(z))

def from_likelihood(p):
    """convert likelyhood to z_score"""
    try:
        return [from_likelihood(x) for x in p]
    except TypeError:
        return st.norm.ppf(1-p/2)


def is_same(a, b):
    """how likely is it that both statistics objects describe the same distribution? Assumes Gaussian distribution and returns the z_score. Use to_likelihood"""
    try:
        ma=a["mean"]
        mb=b["mean"]
        sa=a["std"]
        sb=b["std"]
        na=a["n"]
        nb=b["n"]
    except:
        raise KeyError("a and b must have keys 'mean', 'std', 'n'")

    sa/=np.sqrt(na)
    sb/=np.sqrt(nb)
    s=np.sqrt(sa**2+sb**2)
    if s==0:raise ValueError("std is zero")
    delta=np.abs(ma-mb)
    return delta/s

def similarity_matrix(stats):
    """return a matrix of likelihoods"""
    n=len(stats)
    matrix=np.zeros((n,n))
    
    for i in range(n):
        matrix[i,i]=1
        for j in range(i+1,n):
            matrix[j,i]=matrix[i,j]=is_same(stats[i],stats[j])

    return matrix

if __name__=="__main__":
    print(to_likelihood(0.0))
    print(to_likelihood(1.0))
    print(to_likelihood(2.0))
    print(to_likelihood(3.0))
    print(from_likelihood(to_likelihood(1.234)))













import numpy as np
from sklearn.decomposition import PCA


class projector(object):
    def __init__(self, pca, mu, n_components=2):
        self.pca = pca
        self.mu = mu
        self.n_components = n_components

    def transform(self,X):
        return self.pca.transform(X)

    def inverse_transform(self,X):
        Xhat=np.dot(X,self.pca.components_)
        return Xhat+self.mu






def learn_projection(X, n_components=2):
    """
    Learn a projection matrix that projects the data in X onto a lower dimensional space.
    """
    pca = PCA(n_components=n_components)
    pca.fit(X)
    return projector(pca,np.mean(X,axis=0),n_components=n_components)









from ..data.iterate import iterate

class logic_obj(object):
    def evaluate(self,value):
        return bool(self.getvalue(value))
    def __call__(self,value):
        return self.evaluate(value)
    def __getitem__(self,value):
        return self.evaluate(value)
    
    def getvalue(self,value):
        assert False, f"Only a condition (not {type(self)}) can be evaluated"
        


    def __and__(self,other):
        return logic_and(self,other)
    
    def __or__(self,other):
        return logic_or(self,other)

    def __invert__(self):
        return logic_not(self)


    def __lt__(self,other):
        assert type(other) is int or type(other) is float or type(other) is str or hasattr(other,"getvalue"), "Only numbers can be compared"
        return logic_condition(self, other, "below")

    def __gt__(self,other):
        assert type(other) is int or type(other) is float or type(other) is str or hasattr(other,"getvalue"), "Only numbers can be compared"
        return logic_condition(self, other, "above")

    def __eq__(self,other):
        assert type(other) is int or type(other) is float or type(other) is str or hasattr(other,"getvalue"), "Only numbers can be compared"
        return logic_condition(self, other, "equal")

    def __ne__(self,other):
        assert type(other) is int or type(other) is float or type(other) is str or hasattr(other,"getvalue"), "Only numbers can be compared"
        return logic_condition(self, other, "not_equal")

    def __le__(self,other):
        assert type(other) is int or type(other) is float or type(other) is str or hasattr(other,"getvalue"), "Only numbers can be compared"
        return logic_condition(self, other, "below_or_equal")

    def __ge__(self,other):
        assert type(other) is int or type(other) is float or type(other) is str or hasattr(other,"getvalue"), "Only numbers can be compared"
        return logic_condition(self, other, "above_or_equal")

    def __sub__(self,other):
        assert type(other) is int or type(other) is float or type(other) is str or hasattr(other,"getvalue"), "Only numbers can be compared"
        return logic_condition(self,other,"subtract")

    def __rsub__(self,other):
        assert type(other) is int or type(other) is float or type(other) is str or hasattr(other,"getvalue"), "Only numbers can be compared"
        return logic_condition(other,self,"subtract")

    def __add__(self,other):
        assert type(other) is int or type(other) is float or type(other) is str or hasattr(other,"getvalue"), "Only numbers can be compared"
        return logic_condition(self,other,"add")

    def __radd__(self,other):
        assert type(other) is int or type(other) is float or type(other) is str or hasattr(other,"getvalue"), "Only numbers can be compared"
        return logic_condition(other,self,"add")

    def __mul__(self,other):
        assert type(other) is int or type(other) is float or type(other) is str or hasattr(other,"getvalue"), "Only numbers can be compared"
        return logic_condition(self,other,"multiply")

    def __rmul__(self,other):
        assert type(other) is int or type(other) is float or type(other) is str or hasattr(other,"getvalue"), "Only numbers can be compared"
        return logic_condition(other,self,"multiply")

    def __truediv__(self,other):
        assert type(other) is int or type(other) is float or type(other) is str or hasattr(other,"getvalue"), "Only numbers can be compared"
        return logic_condition(self,other,"divide")

    def __rtruediv__(self,other):
        assert type(other) is int or type(other) is float or type(other) is str or hasattr(other,"getvalue"), "Only numbers can be compared"
        return logic_condition(other,self,"divide")

    def __mod__(self,other):
        assert type(other) is int or type(other) is float or type(other) is str or hasattr(other,"getvalue"), "Only numbers can be compared"
        return logic_condition(self,other,"modulo")

    def __rmod__(self,other):
        assert type(other) is int or type(other) is float or type(other) is str or hasattr(other,"getvalue"), "Only numbers can be compared"
        return logic_condition(other,self,"modulo")

    def __iter__(self):
        return iterate(self)

    def count(self):
        return len(list(iterate(self)))
    
    def __len__(self):
        return self.count()




class logic_condition(logic_obj):
    def __init__(self, a, b, condition="below"):
        self.a = a
        self.b = b
        self.condition = condition

    def getvalue(self, value):
        assert type(value) is dict, "Evalution requires a dictionary of values"
        v1=self.a
        v2=self.b
        def refine(v):
            if hasattr(v,"getvalue"):
                return v.getvalue(value)
            elif v in value.keys():
                return value[v]
            elif type(v) is str or type(v) is int or type(v) is float:
                return v
            else:
                assert False, f"{v}({type(v)}) is not a valid value"
        v1=refine(v1)
        v2=refine(v2)
        #print("comparing",self.condition,self.a,"=",v1,"  ",self.b,"=",v2,"(",type(self.a),type(self.b),")")
        if self.condition == "below":
            return v1 < v2
        elif self.condition == "above":
            return v1 > v2
        elif self.condition == "equal":
            return v1 == v2
        elif self.condition == "not_equal":
            return v1 != v2
        elif self.condition == "below_or_equal":
            return v1 <= v2
        elif self.condition == "above_or_equal":
            return v1 >= v2
        elif self.condition == "subtract":
            return v1 - v2
        elif self.condition == "add":
            return v1 + v2
        elif self.condition == "multiply":
            return v1 * v2
        elif self.condition == "divide":
            return v1 / v2
        elif self.condition == "modulo":
            return v1 % v2
        else:
            assert False, f"Unknown condition {self.condition}"

    def __repr__(self):
        if self.condition == "below":
            return "{} < {}".format(self.a, self.b)
        elif self.condition == "above":
            return "{} > {}".format(self.a, self.b)
        elif self.condition == "equal":
            return "{} == {}".format(self.a, self.b)
        elif self.condition == "not_equal":
            return "{} != {}".format(self.a, self.b)
        elif self.condition == "below_or_equal":
            return "{} <= {}".format(self.a, self.b)
        elif self.condition == "above_or_equal":
            return "{} >= {}".format(self.a, self.b)
        elif self.condition == "subtract":
            return "{} - {}".format(self.a,self.b)
        elif self.condition == "add":
            return "{} + {}".format(self.a,self.b)
        elif self.condition == "multiply":
            return "{} * {}".format(self.a,self.b)
        elif self.condition == "divide":
            return "{} / {}".format(self.a,self.b)
        elif self.condition == "modulo":
            return "{} % {}".format(self.a,self.b)
        else:
            return "Unknown condition"

class logic_and(logic_obj):
    def __init__(self, *args):
        self.args = args

    def getvalue(self, value):
        for arg in self.args:
            if not arg.evaluate(value):
                return 0
        return 1

    def __repr__(self):
        return "("+(" and ".join([str(arg) for arg in self.args]))+")"

class logic_or(logic_obj):
    def __init__(self, *args):
        self.args = args

    def getvalue(self, value):
        for arg in self.args:
            if arg.evaluate(value):
                return 1
        return 0

    def __repr__(self):
        return "("+(" or ".join([str(arg) for arg in self.args]))+")"

class logic_not(logic_obj):
    def __init__(self,obj):
        self.obj=obj

    def getvalue(self,value):
        val=self.obj.evaluate(value)
        return 1-int(val)

    def __repr__(self):
        return "not "+str(self.obj)















class logic_variable(logic_obj):
    def __init__(self, name):
        self.name = name


    def getvalue(self, value):
        if self.name in value.keys():
            return value[self.name]
        else:
            return 0

    def __repr__(self):
        return self.name



def filter_data(data,condition):
    for x in data:
        if condition.evaluate(x):
            yield x








import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import backend as K

import os
import numpy as np
import time

from sklearn.metrics import roc_auc_score

from tqdm import tqdm

from .dean import DEAN

from .densedrop import densedrop


class RandNet(DEAN):
    def __init__(s,*args,pth="./RandNet/",max_epochs=300,alpha=0.5,validation_split=0.1,lr=0.01,**kwargs):
        super().__init__(*args,pth=pth,max_epochs=max_epochs,validation_split=validation_split,lr=lr,**kwargs)
        s.alpha=alpha

    def build_one_model(s,index=0):
        s._seed(index)
        def genlayer(dim,act=s.act):
            return densedrop(dim,activation=act),keras.layers.Dense(dim,activation=act)
        def genlayers(dims):
            layers=[]
            for dim in dims[:-1]:
                layers.append(genlayer(dim))
            layers.append(genlayer(dims[-1],act="linear"))
            return layers
        def genmodel(inp,inp2,layers):
            q=inp
            q2=inp2
            for layer in layers:
                q=layer[0](q)
                q2=layer[1](q2)
            return keras.models.Model(inp,q),keras.models.Model(inp2,q2)

        dim0=int(s.data.shape[1])
        dims=[]
        for i in range(s.depth):
            dims.append(s.alpha**(i+1))
        for i in list(range(s.depth))[::-1]:
            dims.append(s.alpha**i)
        dims=[max([3,int(zw*dim0)]) for zw in dims]


        inp=keras.layers.Input(shape=s.data.shape[1:])
        inp2=keras.layers.Input(shape=s.data.shape[1:])
        layers=genlayers(dims)
        model,model2=genmodel(inp,inp2,layers)

        model.compile(optimizer=keras.optimizers.Adam(lr=s.lr),loss="mse")
        model2.compile(optimizer=keras.optimizers.Adam(lr=s.lr),loss="mse")

        return layers,model,model2



    def train_one(s,index=0):
        pth=s.index_path(index)
        
        data=s.data
        layers,model,model2=s.build_one_model(index)

        t1=time.perf_counter()

        model.fit(data,data,
                  epochs=s.max_epochs,
                  batch_size=s.batch,
                  validation_split=s.validation_split,
                  verbose=s.verbose,
                  shuffle=True)

        t2=time.perf_counter()

        x=data
        px=model.predict(x)
        dx=(px-x)**2
        dx=np.mean(dx,axis=1)
        div=np.std(dx)

        wei=model.get_weights()
        alts=[lay[0].get_matrix() for lay in layers]
        for i,w in enumerate(wei):
            if str(alts[0].shape)==str(w.shape):
                wei[i]=alts.pop(0)
                if len(alts)==0:
                    break
        model2.set_weights(wei)

        model2.save(pth+"model.h5")

        scores=[]
        auc=-1.0
        if not s.test_data is None:
            tx=s.test_data
            p=model2.predict(tx)
            d=(p-tx)**2
            d=np.mean(d,axis=1)
            scores=d
            if not s.test_labels is None:
                auc=roc_auc_score(s.test_labels,scores)

        dt=t2-t1

        if s.verbose:
            print("RandNet:",index,":",dt,":",auc)

        np.savez_compressed(pth+"results.npz",s=scores,auc=auc,dt=dt,div=div)

        return scores/div
    
    def ensemble_merge(s,scores):
        return np.median(np.array(scores)**2 ,axis=0)






import matplotlib.pyplot as plt
import numpy as np

from .projector import learn_projection


from .linedraw import draw

def reducedraw(x,tx,ty,score,border,inc=0.0, hist=True):
    if x.shape[1]>2:
        proj=learn_projection(x,n_components=2)

        q=proj.transform(x)
        tq=proj.transform(tx)
        def new_score(q, i=0):
            x=proj.inverse_transform(q)
            p=score[i](x)
            return p
    else:
        q=x
        tq=tx
        def new_score(q, i=0):
            return score[i](q)

    if not type(score) is list:
        score=[score]


    nscores=[lambda q, i=i: new_score(q, i) for i in range(len(score))]

    minx,maxx,miny,maxy=np.min(q[:,0]),np.max(q[:,0]),np.min(q[:,1]),np.max(q[:,1])
    minx=minx-(maxx-minx)*inc
    maxx=maxx+(maxx-minx)*inc
    miny=miny-(maxy-miny)*inc
    maxy=maxy+(maxy-miny)*inc




    for i,nsc in enumerate(nscores):
        draw(tq,ty,nsc,border,minx=minx,maxx=maxx,miny=miny,maxy=maxy, hist=hist and not i)









import matplotlib.pyplot as plt
import numpy as np

from scipy.optimize import minimize

from tqdm import tqdm


def dualiter(q):
    for i in range(len(q)-1):
        yield q[i], q[i+1]

def draw(samples,truth,func,borders,minx,maxx,miny,maxy, hist=True):

    startingpoints=10#0
    epsilon=0.001
    continuepoints=100
    maxwidth=1000

    linecount=5000
    linesteps=10
    refinement=5

    if type(borders) is not list:
        borders=[borders]


    epsilonx=epsilon*(maxx-minx)
    epsiolny=epsilon*(maxy-miny)

    #print(minx,maxx,miny,maxy)

    sample0=np.array([s for s,t in zip(samples,truth) if t==0])
    sample1=np.array([s for s,t in zip(samples,truth) if t==1])

    if hist:

        if len(sample0)>0 and len(sample1)>0:
            hist0=np.histogram2d(sample0[:,0],sample0[:,1],bins=100,range=[[minx,maxx],[miny,maxy]])[0].T
            hist1=np.histogram2d(sample1[:,0],sample1[:,1],bins=100,range=[[minx,maxx],[miny,maxy]])[0].T
            hist=hist1-hist0
        elif len(sample0)>0:
            hist=np.histogram2d(sample0[:,0],sample0[:,1],bins=100,range=[[minx,maxx],[miny,maxy]])[0].T
        elif len(sample1)>0:
            hist=-np.histogram2d(sample1[:,0],sample1[:,1],bins=100,range=[[minx,maxx],[miny,maxy]])[0].T
        else:
            hist=np.zeros((100,100))

        plt.imshow(hist,extent=[minx,maxx,miny,maxy],origin='lower',cmap="hot")
        plt.grid(False)

    for border in borders:
    
        #plt.hist2d(samples[:,0],samples[:,1],bins=100,range=[[minx,maxx],[miny,maxy]])
    
        #plt.fill(dx,dy,color="red",alpha=0.1)
    
    
        def color(point):
            return (func(point)>border).astype(float)
    
    
        def intermediate_line(x1,x2,y1,y2,cou=100):
            dx,dy=x2-x1,y2-y1
            t=np.arange(0,1,1/cou)
            x=x1+dx*t
            y=y1+dy*t
            q=np.concatenate((np.expand_dims(x,1),np.expand_dims(y,1)),axis=1)
            return q
    
        def random_line(cou=100):
            x1,x2=np.random.uniform(minx,maxx,2)
            y1,y2=np.random.uniform(miny,maxy,2)
            return intermediate_line(x1,x2,y1,y2,cou)
        
        todoo=[random_line(cou=linesteps) for i in range(linecount)]
        hyperbreak=False
        for i in tqdm(range(refinement)):
            if len(todoo)==0:
                print("failed to find anomalousness value")
                hyperbreak=True
                continue
            nextlines=[]
            lines=np.array(todoo)
            lineshape=lines.shape
            lines=lines.reshape(lineshape[0]*lineshape[1],2)
        
        
            colors=color(lines)
            colors=colors.reshape(lineshape[:-1])
            lines=lines.reshape(lineshape)
        
        
            for line,colo in zip(lines,colors):
                for (x1,x2),(c1,c2) in zip(dualiter(line),dualiter(colo)):
                    #print(x1,x2,c1,c2)
                    if c1!=c2:
                        nextlines.append((x1,x2))
            todoo=[intermediate_line(x1[0],x2[0],x1[1],x2[1],cou=linesteps) for x1,x2 in nextlines]
        if hyperbreak:break
        matter=[]
        for x1,x2 in nextlines:
            matter.append((x1+x2)/2)
    
        matter=np.array(matter)
    
    
    
        plt.plot(matter[:,0],matter[:,1],".",alpha=0.3)
    
    
    
    
    



from distutils.core import setup
from setuptools import find_packages

setup(
  name = 'kyano',         
  packages = find_packages(),
  version = '0.7',      
  license='apache-2.0',        
  description = 'Python module providing anomaly detection datasets and tools for evaluating algorithms. Modification of yano, mostly by replacing the datasets, through a kaggle query',   
  author = 'Simon Klüttermann',                   
  author_email = 'Simon.Kluettermann@cs.tu-dortmund.de',      
  url = 'https://github.com/psorus/kano',   
  download_url = 'https://github.com/psorus/kano/archive/v_01.tar.gz',    
  keywords = ['DATA', 'ANOMALY DETECTION', 'OUTLIER DETECTION', 'ML'],   
  install_requires=[            
          'numpy',
          'sklearn',
          'tqdm',
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',      
    'Intended Audience :: Science/Research',      
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: Apache Software License',
    'Programming Language :: Python :: 3',      
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
  ],
)




from .loaddata import allfiles, fileinfo
from .dataset import dataset


def _iterate_all():
    for f in allfiles():
        yield f

def _iterate_condition(condition):
    if callable(condition):
        for f in _iterate_all():
            if condition(fileinfo(f)):
                yield f
    elif type(condition) is dict:
        for f in _iterate_all():
            if not f in condition:continue
            if condition[f]:
                yield f
    elif hasattr(condition, '__iter__'):
        for f in _iterate_all():
            if f in condition:
                yield f
    else:
        try:
            b=bool(condition)
        except:
            raise ValueError('if condition is not defined by metalogic, it must be callable, dict or iterable')
        if b:
            for f in _iterate_all():
                yield f



def _iterate(condition=None):
    if condition is None:
        return _iterate_all()
    else:
        return _iterate_condition(condition)

def iterate(condition=None):
    for f in _iterate(condition):
        yield dataset(f)

def load_dataset(nam):
    #condition=symbols.name==nam
    for d in iterate():
        if d.name()==nam:
            return d
    return None

if __name__=="__main__":
    for f in iterate():
        print(f)







import numpy as np

def seed_once(func, seed=42):

    def wrapper(*args,seed=seed, **kwargs):
        post=np.random.randint(0,2**32-1)
        np.random.seed(seed)
        result = func(*args, **kwargs)
        np.random.seed(post)
        return result
    return wrapper

def lambify(func):

    def wrapper(*args, **kwargs):
        if len(args) > 0 and (hasattr(args[0], '__iter__') and not isinstance(args[0], str)):
            return func(*args, **kwargs)
        else:
            return lambda x: func(x,*args, **kwargs)
    return wrapper

def adaptribute(obj, attr, els=None):
    if hasattr(obj, attr):
        return getattr(obj, attr)
    elif els is not None:
        return els
    else:
        return obj

def adaptricallable(obj, attr, els=None):
    if hasattr(obj, attr):
        return getattr(obj, attr)
    elif els is not None:
        return els
    else:
        return lambda obj=obj: obj

def nannable(func):
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except:
            return np.nan
    return wrapper





from .loaddata import fileinfo, loaddata




class dataset(object):

    def __init__(s, key, x=None, y=None, autoload=False):
        key=str(key)
        s.key = key
        s.x = x
        s.y = y
        s.loaded=not ((x is None) and (y is None))
        if autoload and not s.loaded:
            s.load()



    def info(s):
        return fileinfo(s.key)

    def types(s):
        return s.info()['types']

    def name(s):
        return s.key

    def name_and_feat(s):
        return f"{s.name()}({s.info()['number_of_features']})"

    def load(s):
        s.x,s.y=loaddata(s.key)
        s.loaded=True

    def assert_loaded(s):
        if not s.loaded:
            s.load()

    def getx(s):
        s.assert_loaded()
        return s.x

    def gety(s):
        s.assert_loaded()
        return s.y

    def getxy(s):
        s.assert_loaded()
        return s.x,s.y

    def __str__(s):
        if s.loaded:
            return str(s.key)
        else:
            return "["+str(s.key)+"]"

    def __repr__(s):
        if s.loaded:
            return f"dataset({s.key},autoload=True)"
        else:
            return f"dataset({s.key})"

















from .metalogic import logic_variable


number_of_features = logic_variable("number_of_features")
number_of_samples = logic_variable("number_of_samples")#test+train
categorical = logic_variable("categorical")
numeric = logic_variable("numeric")
nominal = logic_variable("nominal")
name = logic_variable("name")
textual = logic_variable("textual")
time_series = logic_variable("time_series")
twodim = logic_variable("twodim")
threedim = logic_variable("threedim")
image_data = logic_variable("image_data")
image_based = logic_variable("image_based")#no longer an image, but derived from an image
linearly_separable = logic_variable("linearly_separable")
easy_to_separate = logic_variable("easy_to_separate")#in tests most algorithms achieve auc>0.9
hard_to_separate = logic_variable("hard_to_separate")#in tests no algorithm achieves auc>0.9
number_anomalies = logic_variable("number_anomalies")
number_normals = logic_variable("number_normals")
index = logic_variable("index")
fraction_anomalies = logic_variable("fraction_anomalies")#number_anomalies/number_of_samples
quantconst = logic_variable("quantconst")#min options of feature over the numerical features. Only defined if twodim==True
quantmedian = logic_variable("quantmedian")#median options of feature over the numerical features. Only defined if twodim==True
quantisation = logic_variable("quantisation")#quantconst/numberof_samples<0.2

logical_false = logic_variable("logical_false")
logical_true = ~logical_false

size=logic_variable("size")




import os
import sys

from os.path import expanduser

import requests
import json

from .timestamp import timestamp

from functools import lru_cache

import numpy as np

#hosts=["https://raw.githubusercontent.com/psorus/kanodata/main/{fn}"]
hosts=["http://174.138.177.21:2156/{fn}"]

dic_frequency=3600*24


@lru_cache(maxsize=1)
def getbasepth():
    pth=expanduser("~")+"/.kano/data/"
    os.makedirs(pth, exist_ok=True)
    return pth

def calcpath(fn):
    return getbasepth()+fn

def no_file_name(fn):
    if "/" in fn:
        fn=fn[fn.rfind("/")+1:]
    if "." in fn:
        fn=fn[:fn.rfind(".")]
    return fn


def download(fn):
    outp=calcpath(fn)
    for host in hosts:
        url=host.format(fn=fn)
        print(f"Downloading {fn} from {url}")
        r=requests.get(url)
        if r.status_code==200:
            with open(outp, "wb") as f:
                f.write(r.content)
            ages=loadages()
            ages[no_file_name(fn)]=timestamp()
            saveages(ages)
            return
    print(f"Failed to download {fn}")
    exit()

def loadages():
    if os.path.isfile(calcpath("ages.json")):
        return json.load(open(calcpath("ages.json")))
    else:
        return {}

def force_update_dic():
    download("dic.json")

@lru_cache(maxsize=1)
def loaddic():
    assertfile("dic.json",do_check_age=False)
    loct=local_update("dic.json")
    if (loct is None) or (loct<timestamp()-dic_frequency):
        download("dic.json")
    return json.load(open(calcpath("dic.json")))

def fileinfo(fn):
    dic=loaddic()
    fn=no_file_name(fn)
    assert fn in dic.keys(), f"{fn} unknown"
    return dic[fn]

def allfiles():
    dic=loaddic()
    return list(dic.keys())

def countfiles():
    return len(allfiles())

def saveages(ages):
    ages["last"]=timestamp()
    with open(calcpath("ages.json"), "w") as f:
        json.dump(ages, f, indent=2)

def official_update(fn):
    dic=loaddic()
    fn=no_file_name(fn)
    assert fn in dic.keys(), f"{fn} untracked"
    try:
        return dic[fn]["last"]
    except KeyError:
        return 0

def local_update(fn):
    ages=loadages()
    fn=no_file_name(fn)
    if fn in ages.keys():
        return ages[fn]
    else:
        return None


def check_age(fn):
    official=official_update(fn)
    local=local_update(fn)
    if local is None:
        return False
    return local>=official

def assertfile(fn,do_check_age=True):
    if not (os.path.isfile(calcpath(fn)) and (not do_check_age or check_age(fn))):
        download(fn)
    return calcpath(fn)


def loaddata(fn):
    fn=fn+".npz"
    assertfile(fn)
    f=np.load(calcpath(fn))
    return f["x"],f["y"]


if __name__=="__main__":

    #dic=loaddic()

    #print(json.dumps(dic, indent=2))

    x,y=loaddata("cardio")
    print(x.shape,y.shape)









import matplotlib.pyplot as plt
import numpy as np

from .reducedraw import reducedraw



def altitude_plot(func, x,tx,ty, altitude=0.1,inc=0.0, hist=True):
    reducedraw(x,tx,ty,func,altitude,inc=inc, hist=hist)



